/* SmartEar 0.70 by Matt Perry
 * Works for gaim 0.70
 * Plugin to assign different sounds to different buddies and groups.
 */

#include "config.h"
#include "internal.h"

#include "gtkplugin.h"
#include "gtkutils.h"

#include "debug.h"
#include "util.h"
#include "ui.h"
#include "sound.h"
#include "gtkprefs.h"
#include "signals.h"

#ifndef GAIM_PLUGINS
#define GAIM_PLUGINS
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>

#ifdef _WIN32
#include "win32dep.h"
#endif

#define SMARTEAR_PLUGIN_ID "smartear"

#define FIND(s) lookup_widget(config, s)

// an entry in the prefs file to indicate what to play for who and when.
struct sound_event {
    char type; // B,G
    char *name;
    unsigned when; // MSIA
#define SOUNDFLAG_M	1
#define SOUNDFLAG_S	2
#define SOUNDFLAG_I	4
#define SOUNDFLAG_A	8
    char *sound;
};

struct timer_data {
    int event;
    guint id;
};

struct message_data {
    GaimBuddy *buddy;
    time_t last_active;
    time_t last_sent;
    struct timer_data *timer;
};

enum { // Treeview columns
    DATA_COL = 0,
    TYPE_COL = 0,
    NAME_COL,
    ON_MESSAGE_COL,
    ON_SIGNON_COL,
    ON_UNIDLE_COL,
    ON_UNAWAY_COL,
    SOUND_COL,
    NUM_COLS
};

#define DEFAULT_NAME "(Default)"
#define IS_DEFAULT(entry) (strcmp(entry->name, DEFAULT_NAME)==0)
struct sound_event default_entry = {'B', DEFAULT_NAME, SOUNDFLAG_M, ""};

int message_delay = 60;
int focused_quiet = TRUE;
int smartear_timers = FALSE;

GtkWidget *config = NULL;
GtkWidget *file_browse = NULL;
GtkTreeView *treeview = NULL;
GtkTreeSelection *treeselect = NULL;
GtkTreeModel *treemodel = NULL;

struct sound_event *selected_entry = NULL;
GtkTreeIter selected_iter;

GSList *sounds_list = NULL;
GSList *messages_list = NULL;

GtkWidget *create_file_browse(void);
GtkWidget *create_config(void);

static void new_entry(struct sound_event *entry);
static gboolean get_selection(void);
static gboolean set_selection(struct sound_event *entry);
static void update_list(void);
static void populate_list(GtkListStore *store);
static void list_set(GtkListStore *store, GtkTreeIter *iter, struct sound_event *entry);
static void set_sound_entries(void);
static void set_option_entries(void);

static gint sound_cmp(gconstpointer p1, gconstpointer p2);
static struct sound_event *sound_dup(struct sound_event *entry);
static void sound_free(struct sound_event *entry);
static void soundlist_free(void);

static void on_treeselect_changed(GtkTreeSelection *selection, gpointer data);

static void smartear_save(void);
static void smartear_load(void);

static void play_sound_alias(char *sound);

/*** Glade's Support Function ***/

GtkWidget*
lookup_widget                          (GtkWidget       *widget,
                                        const gchar     *widget_name)
{
    GtkWidget *found_widget;
    found_widget = (GtkWidget*) g_object_get_data (G_OBJECT (widget),
	    widget_name);
    if (!found_widget)
	g_warning ("smartear: Widget not found: %s", widget_name);
    return found_widget;
}

/*** GTK Callbacks ***/

void
on_delay_changed                       (GtkEditable     *editable,
                                        gpointer         user_data)
{
    gchar *text = gtk_editable_get_chars(editable, 0, -1);
    message_delay = atoi(text);
    g_free(text);
}

void
on_focus_toggled                       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
    focused_quiet = gtk_toggle_button_get_active(togglebutton);
}

void
on_timer_toggled                       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
    smartear_timers = gtk_toggle_button_get_active(togglebutton);
}

void
on_test_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
    gchar *text =
	gtk_editable_get_chars(GTK_EDITABLE(FIND("sound_entry")), 0, -1);
    play_sound_alias(text);
    g_free(text);
}


void
on_browse_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    if (file_browse)
	return;

    file_browse = create_file_browse();
    if (get_selection())
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_browse),
		selected_entry->sound);
    gtk_widget_show(file_browse);
}

void
on_browse_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
    const char *sound;
    if (!get_selection())
	return;

    sound = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_browse));
    g_free(selected_entry->sound);
    selected_entry->sound = g_strdup(sound);
    update_list();
}

void
on_file_browse_destroy                 (GtkObject       *object,
                                        gpointer         user_data)
{
    file_browse = NULL;
}

void
on_sound_changed                       (GtkEditable     *editable,
                                        gpointer         user_data)
{
    gchar *text;

    if (!get_selection())
	return;

    text = gtk_editable_get_chars(editable, 0, -1);

    g_free(selected_entry->sound);
    selected_entry->sound = text;

    update_list();
}

void
on_name_changed                        (GtkEditable     *editable,
                                        gpointer         user_data)
{
    gchar *text;

    if (!get_selection())
	return;

    if (IS_DEFAULT(selected_entry)) {
	g_signal_handlers_block_by_func(G_OBJECT(editable), on_name_changed, 0);
	gtk_entry_set_text(GTK_ENTRY(editable), default_entry.name);
	g_signal_handlers_unblock_by_func(G_OBJECT(editable), on_name_changed, 0);
	return;
    }
    text = gtk_editable_get_chars(editable, 0, -1);

    g_free(selected_entry->name);
    selected_entry->name = text;

    update_list();
}

void on_type_option_activate(GtkMenuItem *menuitem, gpointer user_data);

void
on_buddy_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    on_type_option_activate(menuitem, GINT_TO_POINTER('B'));
}

void
on_group_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    on_type_option_activate(menuitem, GINT_TO_POINTER('G'));
}

// stupid Glade2 won't let me set user data, so I have to go around it
void on_type_option_activate(GtkMenuItem *menuitem, gpointer user_data)
{
    gchar type = GPOINTER_TO_INT(user_data);

    if (!get_selection())
	return;

    if (IS_DEFAULT(selected_entry)) {
	g_signal_handlers_block_by_func(G_OBJECT(menuitem), on_buddy_activate, 0);
	gtk_option_menu_set_history(GTK_OPTION_MENU(FIND("type_option")), 0);
	g_signal_handlers_unblock_by_func(G_OBJECT(menuitem), on_buddy_activate, 0);
	return;
    }

    selected_entry->type = type;
    update_list();
}


void
on_new_clicked                         (GtkButton       *button,
                                        gpointer         user_data)
{
    struct sound_event *entry, tmp;

    tmp = default_entry;
    tmp.name = "";
    entry = sound_dup(&tmp);
    new_entry(entry);
}


void
on_delete_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    GtkTreeIter iter = selected_iter;

    if (!get_selection())
	return;

    if (IS_DEFAULT(selected_entry))
	return;

    sounds_list = g_slist_remove(sounds_list, selected_entry);
    sound_free(selected_entry);

    if (!gtk_tree_model_iter_next(treemodel, &iter)) {
	GtkTreePath *path = gtk_tree_model_get_path(treemodel, &selected_iter);
	gtk_tree_path_prev(path);
	gtk_tree_model_get_iter(treemodel, &iter, path);
	gtk_tree_path_free(path);
    }
    g_signal_handlers_block_by_func(G_OBJECT(treeselect), on_treeselect_changed, 0);
    gtk_list_store_remove(GTK_LIST_STORE(treemodel), &selected_iter);
    gtk_tree_selection_select_iter(treeselect, &iter);
    g_signal_handlers_unblock_by_func(G_OBJECT(treeselect), on_treeselect_changed, 0);

    get_selection();
    update_list();
}


void on_cell_edited(GtkCellRendererText *cell, gchar *path, gchar *text, gpointer data)
{
    GtkTreeIter iter;
    gint col = GPOINTER_TO_INT(data);
    struct sound_event *entry;

    if (!gtk_tree_model_get_iter_from_string(treemodel, &iter, path))
	return;

    gtk_tree_model_get(treemodel, &iter, DATA_COL, &entry, -1);
    switch (col) {
    case NAME_COL:
	if (IS_DEFAULT(entry))
	    return;
	g_free(entry->name);
	entry->name = g_strdup(text);
	update_list();
	break;
    case SOUND_COL:
	g_free(entry->sound);
	{
	    char buf[BUF_LEN];

	    if (strncmp(text, "~", 1) == 0 && gaim_home_dir())
		g_snprintf(buf, sizeof(buf), "%s%s", gaim_home_dir(), text+1);
	    else
		g_snprintf(buf, sizeof(buf), "%s", text);
	    entry->sound = g_strdup(buf);
	}
	update_list();
	break;
    }
}

void on_cell_toggled(GtkCellRendererToggle *cell, gchar *path, gpointer data)
{       
    GtkTreeIter iter;
    gint bit = 0;
    gint col = GPOINTER_TO_INT(data);
    struct sound_event *entry;

    if (!gtk_tree_model_get_iter_from_string(treemodel, &iter, path))
	return;

    gtk_tree_model_get(treemodel, &iter, DATA_COL, &entry, -1);
    switch (col) {
    case ON_MESSAGE_COL:	bit = SOUNDFLAG_M; break;
    case ON_SIGNON_COL:		bit = SOUNDFLAG_S; break;
    case ON_UNIDLE_COL:		bit = SOUNDFLAG_I; break;
    case ON_UNAWAY_COL:		bit = SOUNDFLAG_A; break;
    }
    entry->when ^= bit;
}

void on_treeselect_changed(GtkTreeSelection *selection, gpointer data)
{
    get_selection();
    update_list();
}


void
on_save_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
    smartear_save();
}

void
on_revert_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    selected_entry = NULL;
    soundlist_free();
    smartear_load();
    gtk_list_store_clear(GTK_LIST_STORE(treemodel));
    populate_list(GTK_LIST_STORE(treemodel));
    update_list();
    set_option_entries();
}

#include "smartear-interface.c"

/*** Util Functions ***/

static void new_entry(struct sound_event *entry)
{
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeViewColumn *name_column;

    if (entry->name && entry->name[0] && set_selection(entry)) {
	sound_free(entry);
	return;
    }

    sounds_list = g_slist_append(sounds_list, entry);
    if (get_selection())
	gtk_list_store_insert_after(GTK_LIST_STORE(treemodel), &iter, &selected_iter);
    else
	gtk_list_store_append(GTK_LIST_STORE(treemodel), &iter);

    path = gtk_tree_model_get_path(treemodel, &iter);
    list_set(GTK_LIST_STORE(treemodel), &iter, entry);

    g_signal_handlers_block_by_func(G_OBJECT(treeselect), on_treeselect_changed, 0);
    path = gtk_tree_model_get_path(treemodel, &iter);
    name_column = gtk_tree_view_get_column(treeview, NAME_COL);
    if (entry->name && entry->name[0])
	gtk_tree_view_set_cursor(treeview, path, 0, FALSE);
    else
	gtk_tree_view_set_cursor(treeview, path, name_column, TRUE);
    gtk_tree_path_free(path);
    g_signal_handlers_unblock_by_func(G_OBJECT(treeselect), on_treeselect_changed, 0);

    get_selection();
    set_sound_entries();
}

static gboolean get_selection(void)
{
    if (!gtk_tree_selection_get_selected(treeselect, &treemodel, &selected_iter))
	return FALSE;

    gtk_tree_model_get(treemodel, &selected_iter, DATA_COL, &selected_entry, -1);
    return TRUE;
}

static gboolean set_selection(struct sound_event *entry)
{
    GtkTreeIter iter;

    if (!gtk_tree_model_get_iter_first(treemodel, &iter))
	return FALSE;

    while (1) {
	struct sound_event *entry2;
	gtk_tree_model_get(treemodel, &iter, DATA_COL, &entry2, -1);
	if (sound_cmp(entry, entry2) == 0) {
	    gtk_tree_selection_select_iter(treeselect, &iter);
	    return TRUE;
	}
	if (!gtk_tree_model_iter_next(treemodel, &iter))
	    break;
    }
    return FALSE;
}

static void list_set(GtkListStore *store, GtkTreeIter *iter,
	struct sound_event *entry)
{
    gtk_list_store_set(store, iter, DATA_COL, entry, -1);
}

static char *my_normalize(const char *input)
{
    static char buf[BUF_LEN];
    char *str, *beg;
    int i=0;

    g_return_val_if_fail((input != NULL), NULL);

    beg = str = g_strdup(input);

    while (*str && i <= BUF_LEN) {
	if (*str != ' ')
	    buf[i++] = *str;
	str++;
    }
    buf[i] = '\0';
    g_free(beg);

    return buf;

}

static struct sound_event *sound_dup(struct sound_event *entry)
{
    struct sound_event *edup = g_new(struct sound_event, 1);

    edup->type = entry->type;
    edup->name = g_strdup(my_normalize(entry->name));
    edup->when = entry->when;
    edup->sound = g_strdup(entry->sound);

    return edup;
}

static void sound_free(struct sound_event *entry)
{
    g_free(entry->name);
    g_free(entry->sound);
    g_free(entry);
}

static void soundlist_free(void)
{
    GSList *lp;
    for (lp = sounds_list; lp; lp = g_slist_next(lp))
	sound_free((struct sound_event*)lp->data);
    if (sounds_list) {
	g_slist_free(sounds_list);
	sounds_list = 0;
    }
}


static gint sound_cmp(gconstpointer p1, gconstpointer p2)
{
    struct sound_event *entry1 = (struct sound_event*)p1;
    struct sound_event *entry2 = (struct sound_event*)p2;

    if (!entry1)
	return -1;
    if (!entry2)
	return 1;

    // only compare types if both are nonzero
    if (entry1->type != entry2->type &&
	    entry1->type != 0 && entry2->type != 0)
	return (entry1->type - entry2->type);
    return (g_strcasecmp(entry1->name, entry2->name));
}


static void set_sound_entries(void)
{
    GtkEntry *name_entry = GTK_ENTRY(FIND("name_entry"));
    GtkEntry *sound_entry = GTK_ENTRY(FIND("sound_entry"));
    GtkWidget *item_buddy = FIND("item_buddy");
    GtkWidget *item_group = FIND("item_group");
    struct sound_event dummy, *entry;

    dummy.type = 'B';
    dummy.name = "";
    dummy.sound = "";

    if (!(entry = selected_entry))
	entry = &dummy;

    g_signal_handlers_block_by_func(G_OBJECT(name_entry), on_name_changed, 0);
    g_signal_handlers_block_by_func(G_OBJECT(sound_entry), on_sound_changed, 0);
    g_signal_handlers_block_by_func(G_OBJECT(item_buddy), on_buddy_activate, 0);
    g_signal_handlers_block_by_func(G_OBJECT(item_group), on_group_activate, 0);

    gtk_entry_set_text(name_entry, entry->name);
    gtk_entry_set_text(sound_entry, entry->sound);
    gtk_option_menu_set_history(GTK_OPTION_MENU(FIND("type_option")),
	    (entry->type == 'B') ? 0 : 1);

    g_signal_handlers_unblock_by_func(G_OBJECT(name_entry), on_name_changed, 0);
    g_signal_handlers_unblock_by_func(G_OBJECT(sound_entry), on_sound_changed, 0);
    g_signal_handlers_unblock_by_func(G_OBJECT(item_buddy), on_buddy_activate, 0);
    g_signal_handlers_unblock_by_func(G_OBJECT(item_group), on_group_activate, 0);
}

static void set_option_entries(void)
{
    GtkSpinButton *delay_spin = GTK_SPIN_BUTTON(FIND("delay_spin"));
    GtkToggleButton *focus_but = GTK_TOGGLE_BUTTON(FIND("focus_but"));
    GtkToggleButton *timer_but = GTK_TOGGLE_BUTTON(FIND("timer_but"));

    g_signal_handlers_block_by_func(G_OBJECT(delay_spin), on_delay_changed, 0);
    g_signal_handlers_block_by_func(G_OBJECT(focus_but), on_focus_toggled, 0);
    g_signal_handlers_block_by_func(G_OBJECT(timer_but), on_timer_toggled, 0);

    gtk_spin_button_set_value(delay_spin, (gdouble)message_delay);
    gtk_toggle_button_set_active(focus_but, focused_quiet);
    gtk_toggle_button_set_active(timer_but, smartear_timers);

    g_signal_handlers_unblock_by_func(G_OBJECT(delay_spin), on_delay_changed, 0);
    g_signal_handlers_unblock_by_func(G_OBJECT(focus_but), on_focus_toggled, 0);
    g_signal_handlers_unblock_by_func(G_OBJECT(timer_but), on_timer_toggled, 0);
}

static void update_list(void)
{
    set_sound_entries();
    if (selected_entry)
	list_set(GTK_LIST_STORE(treemodel), &selected_iter, selected_entry);
}

/*** Init Functions ***/

static void populate_list(GtkListStore *store)
{
    GtkTreeIter iter;
    GSList *lp;

    sounds_list = g_slist_sort(sounds_list, (GCompareFunc)sound_cmp);

    for (lp = sounds_list; lp; lp = g_slist_next(lp)) {
	struct sound_event *entry = (struct sound_event*)lp->data;
	gtk_list_store_append(store, &iter);
	list_set(store, &iter, entry);
	if (entry == selected_entry)
	    selected_iter = iter;
    }
}

static void render_type(GtkTreeViewColumn *column, GtkCellRenderer *cell,
    GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    struct sound_event *entry;
    gtk_tree_model_get(model, iter, DATA_COL, &entry, -1);
    if (IS_DEFAULT(entry))
	g_object_set(cell, "text", "", NULL);
    else
	g_object_set(cell, "text", entry->type == 'B' ? "Buddy" : "Group", NULL);
}

static void render_name(GtkTreeViewColumn *column, GtkCellRenderer *cell,
    GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    struct sound_event *entry;
    gtk_tree_model_get(model, iter, DATA_COL, &entry, -1);
    g_object_set(cell, "text", entry->name, NULL);
}

static void render_when(GtkTreeViewColumn *column, GtkCellRenderer *cell,
    GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    struct sound_event *entry;
    gint bit = GPOINTER_TO_INT(data);
    gtk_tree_model_get(model, iter, DATA_COL, &entry, -1);
    g_object_set(cell, "active", (entry->when & bit) ? TRUE : FALSE, NULL);
}

static void render_sound(GtkTreeViewColumn *column, GtkCellRenderer *cell,
    GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    struct sound_event *entry;
    gtk_tree_model_get(model, iter, DATA_COL, &entry, -1);

    {
	const char *home = gaim_home_dir();
	char buf[BUF_LEN];

	if (home && strncmp(entry->sound, home, strlen(home)) == 0)
	    g_snprintf(buf, sizeof(buf), "~%s", entry->sound + strlen(home));
	else
	    g_snprintf(buf, sizeof(buf), "%s", entry->sound);
	g_object_set(cell, "text", buf, NULL);
    }
}

static gint list_cmp(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer d)
{
    struct sound_event *entry1, *entry2;

    gtk_tree_model_get(model, a, DATA_COL, &entry1, -1);
    gtk_tree_model_get(model, b, DATA_COL, &entry2, -1);

    return sound_cmp(entry1, entry2);
}

static void setup_list(void)
{
    GtkListStore *store;
    GtkCellRenderer *cell;

    treeview = GTK_TREE_VIEW(FIND("treeview"));

    store = gtk_list_store_new(1, G_TYPE_POINTER);
    populate_list(store);

    gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(store));
    g_object_unref(G_OBJECT(store)); // treeview has it's own copy.
    treemodel = gtk_tree_view_get_model(treeview);
    gtk_tree_sortable_set_default_sort_func(GTK_TREE_SORTABLE(treemodel),
	    (GtkTreeIterCompareFunc)&list_cmp, 0, 0);
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(treemodel),
	    GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID, GTK_SORT_ASCENDING);

    treeselect = gtk_tree_view_get_selection(treeview);
    gtk_tree_selection_set_mode(treeselect, GTK_SELECTION_SINGLE);
    g_signal_connect(G_OBJECT(treeselect), "changed",
	    G_CALLBACK(on_treeselect_changed), NULL);

    // Type column
    cell = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_data_func(treeview, -1, "Type",
	    cell, render_type, 0, 0);

    // Name column
    cell = gtk_cell_renderer_text_new();
    g_signal_connect(G_OBJECT(cell), "edited",
	    G_CALLBACK(on_cell_edited), GINT_TO_POINTER(NAME_COL));
    g_object_set(G_OBJECT(cell), "editable", TRUE, NULL);
    gtk_tree_view_insert_column_with_data_func(treeview, -1, "Name",
	    cell, render_name, 0, 0);

    // 4 When columns
    cell = gtk_cell_renderer_toggle_new();
    g_signal_connect(G_OBJECT(cell), "toggled",
	    G_CALLBACK(on_cell_toggled), GINT_TO_POINTER(ON_MESSAGE_COL));
    gtk_tree_view_insert_column_with_data_func(treeview, -1, "On Message",
	    cell, render_when, GINT_TO_POINTER(SOUNDFLAG_M), 0);

    cell = gtk_cell_renderer_toggle_new();
    g_signal_connect(G_OBJECT(cell), "toggled",
	    G_CALLBACK(on_cell_toggled), GINT_TO_POINTER(ON_SIGNON_COL));
    gtk_tree_view_insert_column_with_data_func(treeview, -1, "On Signon",
	    cell, render_when, GINT_TO_POINTER(SOUNDFLAG_S), 0);

    cell = gtk_cell_renderer_toggle_new();
    g_signal_connect(G_OBJECT(cell), "toggled",
	    G_CALLBACK(on_cell_toggled), GINT_TO_POINTER(ON_UNIDLE_COL));
    gtk_tree_view_insert_column_with_data_func(treeview, -1, "On Unidle",
	    cell, render_when, GINT_TO_POINTER(SOUNDFLAG_I), 0);

    cell = gtk_cell_renderer_toggle_new();
    g_signal_connect(G_OBJECT(cell), "toggled",
	    G_CALLBACK(on_cell_toggled), GINT_TO_POINTER(ON_UNAWAY_COL));
    gtk_tree_view_insert_column_with_data_func(treeview, -1, "On Unaway",
	    cell, render_when, GINT_TO_POINTER(SOUNDFLAG_A), 0);

    // Sound column
    cell = gtk_cell_renderer_text_new();
    g_signal_connect(G_OBJECT(cell), "edited",
	    G_CALLBACK(on_cell_edited), GINT_TO_POINTER(SOUND_COL));
    g_object_set(G_OBJECT(cell), "editable", TRUE, NULL);
    gtk_tree_view_insert_column_with_data_func(treeview, -1, "Sound to Play",
	    cell, render_sound, 0, 0);
}

static void smartear_save(void)
{
    char file[BUF_LEN] = "smartear.rc - you have no home dir";
    FILE *fp = NULL;
    GSList *lp;
    struct sound_event *entry;

    if (gaim_home_dir()) {
	g_snprintf(file, sizeof(file), "%s/.gaim/smartear.rc", gaim_home_dir());
	fp = fopen(file, "w");
    }

    if (fp == NULL) {
	g_warning("couldn't open %s\n", file);
	return;
    }

    fprintf(fp, "smartear_timers %d\n", smartear_timers);
    fprintf(fp, "delay %d\n", message_delay);
    fprintf(fp, "focused_quiet %d\n", focused_quiet);

    for (lp = sounds_list; lp; lp = g_slist_next(lp)) {
	entry = (struct sound_event*)lp->data;
	fprintf(fp, "%c {%s} {", entry->type, entry->name);
	if (entry->when & SOUNDFLAG_M) fputc('M', fp);
	if (entry->when & SOUNDFLAG_S) fputc('S', fp);
	if (entry->when & SOUNDFLAG_I) fputc('I', fp);
	if (entry->when & SOUNDFLAG_A) fputc('A', fp);
	fprintf(fp, "} {%s}\n", entry->sound);
    }

    fclose(fp);
}

static void smartear_load(void)
{
    char file[BUF_LEN] = "smartear.rc - you have no home dir";
    FILE *fp = NULL;
    struct sound_event *entry;
    char buf[BUF_LEN];
    char name[BUF_LEN];
    char flags[BUF_LEN];
    char type;
    gboolean has_default = FALSE;

    if (gaim_home_dir()) {
	g_snprintf(file, sizeof(file), "%s/.gaim/smartear.rc", gaim_home_dir());
	fp = fopen(file, "r");
    }

    if (fp == NULL) {
	g_warning("smartear: couldn't open %s\n", file);
	sounds_list = g_slist_append(sounds_list, sound_dup(&default_entry));
	return;
    }

    while ((fgets(buf, sizeof(buf), fp))) {
	int tmp;
	if (sscanf(buf, "smarterear_timers %d", &tmp) == 1)
	    smartear_timers = tmp;
	else if (sscanf(buf, "smartear_timers %d", &tmp) == 1)
	    smartear_timers = tmp;
	else if (sscanf(buf, "delay %d", &tmp) == 1)
	    message_delay = tmp;
	else if (sscanf(buf, "focused_quiet %d", &tmp) == 1)
	    focused_quiet = tmp;
	else if (sscanf(buf, "%c {%[^}]} {%[^}]} {%[^}]}", &type, name, flags, file) == 4) {
	    if (type != 'G' && type != 'B') {
		g_warning("smartear: rc no such type: %c", type);
		continue;
	    }

	    entry = g_new(struct sound_event, 1);
	    entry->type = type;
	    entry->name = g_strdup(my_normalize(name));
	    entry->when = 0;
	    if (strchr(flags, 'M'))	entry->when |= SOUNDFLAG_M;
	    if (strchr(flags, 'S'))	entry->when |= SOUNDFLAG_S;
	    if (strchr(flags, 'I'))	entry->when |= SOUNDFLAG_I;
	    if (strchr(flags, 'A'))	entry->when |= SOUNDFLAG_A;
	    entry->sound = g_strdup(file);

	    sounds_list = g_slist_append(sounds_list, entry);

	    if (IS_DEFAULT(entry)) {
		default_entry = *entry;
		has_default = TRUE;
	    }
	}
	else
	    g_warning("smartear: rc syntax error: %s", buf);
    }
    if (!has_default)
	sounds_list = g_slist_append(sounds_list, sound_dup(&default_entry));

    fclose(fp);
}

/*** Gaim callbacks ***/

static struct message_data *find_message_by_name(const char *pname)
{
    GSList *lp;
    struct message_data *msg = NULL;
    char *name = g_strdup(my_normalize(pname));

    for (lp = messages_list; lp; lp = g_slist_next(lp)) {
	msg = (struct message_data*)lp->data;
	if (msg->buddy
		&& g_strcasecmp(name, my_normalize(msg->buddy->name)) == 0)
	    break;
    }

    g_free(name);

    if (lp)
	return msg;
    return 0;
}

static void message_timer_remove(struct message_data *msg)
{
    if (msg && msg->timer) {
	g_source_remove(msg->timer->id);
	g_free(msg->timer);
	msg->timer = NULL;
    }
}

static void message_free(struct message_data *msg)
{
    message_timer_remove(msg);
    g_free(msg);
}

static void messagelist_free()
{
    GSList *lp;
    for (lp = messages_list; lp; lp = g_slist_next(lp))
	message_free((struct message_data*)lp->data);
    if (messages_list) {
	g_slist_free(messages_list);
	messages_list = NULL;
    }
}

#if 0
// Yep. It was a dangerous hack.  Broxx0red as of 0.65

G_MODULE_IMPORT GtkTreeStore *prefstree; // gaim global
G_MODULE_IMPORT GtkWidget *tree_v; // gaim global

// This is probably a dangerous hack.  Works as of gaim 0.62
static void on_smartear_clicked(GtkButton *button, gpointer data)
{
    GaimConversation *conv = (gpointer)data;
    GtkTreeIter iter;
    GtkTreeSelection *treeselect;

    // select smartear's config page

    gaim_gtk_prefs_show();

    if (!gtk_tree_model_get_iter_first(GTK_TREE_MODEL(prefstree), &iter))
	return;

    treeselect = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_v));
    while (1) {
	char *title;
	gtk_tree_model_get(GTK_TREE_MODEL(prefstree), &iter, 1, &title, -1);
	if (strcmp(title, "Plugins") == 0) {
	    GtkTreeIter parent = iter;
	    if (!gtk_tree_model_iter_children(GTK_TREE_MODEL(prefstree), &iter, &parent))
		return;
	    g_free(title);
	    continue;
	}
	if (strcmp(title, "SmartEar") == 0) {
	    gtk_tree_selection_select_iter(treeselect, &iter);
	    g_free(title);
	    break;
	}
	g_free(title);
	if (!gtk_tree_model_iter_next(GTK_TREE_MODEL(prefstree), &iter))
	    return;
    }

    // ok, it's selected. now fill in the entry.

    {
	struct sound_event tmp, *entry;

	tmp = default_entry;
	tmp.name = (char*)gaim_conversation_get_name(conv);
	tmp.type = 'B';
	entry = sound_dup(&tmp);

	new_entry(entry);
    }
}
#else
static void on_smartear_clicked(GtkButton *button, gpointer data)
{
    gaim_gtk_prefs_show();
}
#endif

#define SMARTEAR_BUTTON_KEY "SmartEar"
#define CONV_GET_BUTTON(conv)\
    gaim_conversation_get_data(conv, SMARTEAR_BUTTON_KEY)
#define CONV_SET_BUTTON(conv, button)\
    gaim_conversation_set_data(conv, SMARTEAR_BUTTON_KEY, button)
#define CONV_REM_BUTTON(conv)\
    g_hash_table_remove(conv->data, SMARTEAR_BUTTON_KEY)

// add a button to the conv's IM window if it's not already there
static void add_smartear_button(GaimConversation *conv, gpointer data)
{
    GaimGtkConversation *gtkconv = conv ? GAIM_GTK_CONVERSATION(conv) : NULL;
    GtkWidget *button;

    if (!gtkconv)
	return;
    if (CONV_GET_BUTTON(conv))
	return;

    button = gaim_pixbuf_button_from_stock("SmartEar",
	    GTK_STOCK_PREFERENCES, GAIM_BUTTON_VERTICAL);
    button = gaim_gtk_change_text("SmartEar", button,
	    GTK_STOCK_PREFERENCES, GAIM_BUTTON_VERTICAL);
    gtk_tooltips_set_tip(gtkconv->tooltips, button, "Add buddy to SmartEar", NULL);
    g_signal_connect(G_OBJECT(button), "clicked",
	    G_CALLBACK(on_smartear_clicked), conv);
    gtk_box_pack_start(GTK_BOX(gtkconv->bbox), button, FALSE, FALSE, 0);
    gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);

    gtk_widget_show(button);

    CONV_SET_BUTTON(conv, button);
}

// get rid of the smartear button
static void rem_smartear_button(GaimConversation *conv, gpointer data)
{
    GtkWidget *button = GTK_WIDGET(CONV_GET_BUTTON(conv));
    if (button) {
	gtk_widget_hide(button);
	gtk_widget_destroy(button);
	CONV_REM_BUTTON(conv);
    }
}

static void play_sound_alias(char *sound)
{
    struct sound_event tmp, *entry;
    GSList *lp;

    if (!sound || !*sound)
	return;

    // sound aliases: mostly so you can put (Default) for sound, and it'll
    // play that one
    tmp.type = 0;
    tmp.name = sound;
    if ((lp = g_slist_find_custom(sounds_list, &tmp, (GCompareFunc)sound_cmp))) {
	entry = (struct sound_event*)lp->data;
	play_sound_alias(entry->sound);
    }
    else {
	gaim_sound_play_file(sound);
    }
}

void play_matching_sound(GaimBuddy *buddy, int event)
{
    GSList *lp;
    struct sound_event *entry;
    char *sound = NULL;
    char *name = buddy ? g_strdup(my_normalize(buddy->name)) : NULL;
    GaimGroup *g = buddy ? gaim_find_buddys_group(buddy) : NULL;

    for (lp = sounds_list; lp; lp = g_slist_next(lp)) {
	entry = (struct sound_event*)lp->data;
	if ((entry->when & event) == 0)
	    continue;

	if (entry->type == 'B' && name
		&& g_strcasecmp(name, entry->name) == 0) {
	    sound = entry->sound;
	    break; // found a buddy match.. this takes precedence, so stop
	}
	if (entry->type == 'G' && g
		&& g_strcasecmp(my_normalize(g->name), entry->name) == 0) {
	    sound = entry->sound; // keep going... buddy overrides group
	}
	if (IS_DEFAULT(entry)) {
	    if (!sound)
		sound = entry->sound; // keep going.. other 2 override default
	}
    }
    if (sound) {
	gaim_debug(GAIM_DEBUG_INFO, "smartear", "found %s for %s on event %d\n", sound, name, event);
	play_sound_alias(sound);
    }
    g_free(name);
}

static gint play_sound_timer_hook(gpointer data)
{
    struct message_data *msg = (struct message_data *)data;

    if(!msg || !msg->timer) {
        g_warning("smartear: timer called without being active!\n");
        return FALSE;
    }

    play_matching_sound(msg->buddy, msg->timer->event);

    g_free(msg->timer);
    msg->timer = NULL;

    return FALSE;
}

gboolean on_im_recv(GaimAccount *account, char **who, char **what, gint32 *flags, void *junk)
{
    struct message_data *msg;
    time_t now = time(0);
    GaimConversation *conv = gaim_find_conversation(*who);
    GaimGtkConversation *gtkconv = conv ? GAIM_GTK_CONVERSATION(conv) : NULL;

    // FIXME: message list is never trimmed. However, if we DO trim it,
    // we need to consider race conditions on the msg field of the timer_hook

    if ((msg = find_message_by_name(*who))) {
	// only install a timer if there's no active timers and we've just 
	// sent a message. If we haven't just sent a message, then 
	// a beep will go off anyways when the IM arrives.
	if(smartear_timers && !msg->timer
		&& msg->last_sent + message_delay > now) {
	    msg->timer = g_new0(struct timer_data, 1);
	    msg->timer->event = SOUNDFLAG_M;
	    msg->timer->id = g_timeout_add(message_delay*1000,
		    play_sound_timer_hook, msg);
	}

	if (msg->last_active + message_delay > now && conv)
	    return FALSE; // hasn't been delay seconds yet
    }
    else {
	msg = g_new0(struct message_data, 1);
	msg->buddy = gaim_find_buddy(account, *who);
	msg->timer = NULL;

	messages_list = g_slist_append(messages_list, msg);
    }

    msg->last_active = now;

    if (focused_quiet && gtkconv && GTK_WIDGET_HAS_FOCUS(gtkconv->entry))
	return FALSE; // don't play sounds for the focused convo
    if (gtkconv && !gtkconv->make_sound)
	return FALSE;

    play_matching_sound(msg->buddy, SOUNDFLAG_M);

    return FALSE;
}

void on_im_send(GaimAccount *account, char *who, char *what, void *junk)
{
    struct message_data *msg;
    time_t now = time(0);

    if ((msg = find_message_by_name(who))) {
	message_timer_remove(msg);
    }
    else {
	msg = g_new0(struct message_data, 1);
	msg->buddy = gaim_find_buddy(account, who);
        msg->timer = NULL;

	messages_list = g_slist_append(messages_list, msg);
    }

    msg->last_active = now;
    msg->last_sent = now;
}

void on_buddy_signon(GaimBuddy *buddy, void *data)
{
    play_matching_sound(buddy, SOUNDFLAG_S);
}

void on_buddy_back(GaimBuddy *buddy, void *data)
{
    play_matching_sound(buddy, SOUNDFLAG_A);
}

void on_buddy_unidle(GaimBuddy *buddy, void *data)
{
    play_matching_sound(buddy, SOUNDFLAG_I);
}

void on_signoff(GaimConnection *gc, void *m)
{
    messagelist_free();
}

static void on_new_conv(GaimConversation *conv, void *data)
{
    add_smartear_button(conv, 0);
}

static void on_del_conv(GaimConversation *conv, void *data)
{
    const char *name = gaim_conversation_get_name(conv);
    struct message_data *msg;

    if ((msg = find_message_by_name(name))) {
	message_timer_remove(msg);
    }
    // no need to remove the smartbutton, it will get destroyed by gaim
}

static gboolean gaim_plugin_init(GaimPlugin *plugin)
{
    void *blist_handle = gaim_blist_get_handle();
    void *conv_handle = gaim_conversations_get_handle();

    config = NULL;

    smartear_load();

    gaim_signal_connect(blist_handle, "buddy-signed-on",
	    plugin, GAIM_CALLBACK(on_buddy_signon), NULL);
    gaim_signal_connect(blist_handle, "buddy-back",
	    plugin, GAIM_CALLBACK(on_buddy_back), NULL);
    gaim_signal_connect(blist_handle, "buddy-unidle",
	    plugin, GAIM_CALLBACK(on_buddy_unidle), NULL);

    gaim_signal_connect(conv_handle, "received-im-msg",
	    plugin, GAIM_CALLBACK(on_im_recv), NULL);
    gaim_signal_connect(conv_handle, "sent-im-msg",
	    plugin, GAIM_CALLBACK(on_im_send), NULL);
    gaim_signal_connect(conv_handle, "conversation-created",
	    plugin, GAIM_CALLBACK(on_new_conv), NULL);
    gaim_signal_connect(conv_handle, "deleting-conversation",
	    plugin, GAIM_CALLBACK(on_del_conv), NULL);

    gaim_signal_connect(gaim_connections_get_handle(), "signed-off",
	    plugin, GAIM_CALLBACK(on_signoff), NULL);

    g_list_foreach(gaim_get_conversations(), (GFunc)add_smartear_button, 0);

    return TRUE;
}

static gboolean gaim_plugin_remove(GaimPlugin *h)
{
    config = NULL;

    soundlist_free();
    messagelist_free();

    g_list_foreach(gaim_get_conversations(), (GFunc)rem_smartear_button, 0);

    return TRUE;
}

static GtkWidget *gaim_plugin_config_gtk(GaimPlugin *plugin)
{
    config = create_config();

    setup_list();
    set_sound_entries();
    set_option_entries();

    gtk_widget_show_all(config);

    return config;
}

static GaimGtkPluginUiInfo ui_info = {
    gaim_plugin_config_gtk
};

static GaimPluginInfo info = {
    2,
    GAIM_PLUGIN_STANDARD,
    GAIM_GTK_PLUGIN_TYPE,
    0,
    NULL,
    GAIM_PRIORITY_DEFAULT,

    SMARTEAR_PLUGIN_ID,
    N_("SmartEar"),
    VERSION,
    N_("Assign different sounds to different buddies and groups"),
    N_("Tired of rushing to the computer when you hear an IM noise, "
"expecting an urgent message, only to find out it's your jerk "
"cousin with news about his driveway basketball game victory?  "
"Sick of being interrupted by people you don't want to "
"talk to?  You need this plugin.  With it, you can assign different "
"sounds to play for different buddies or whole groups of buddies.  "
"SmartEar allows you to opt to play sounds when a buddy sends you an "
"IM, signs on, returns from away or idle, or any combination of these, so "
"you'll know by the sound what the important people are doing.  "
"You'll never have to talk to your relatives-in-law again."),
    "Matt Perry &lt;guy@somewhere.fscked.org>",
    "http://somewhere.fscked.org/smartear/",

    gaim_plugin_init, // load
    gaim_plugin_remove, // unload
    NULL, // destroy

    &ui_info,
    NULL
};

static void __init_plugin(GaimPlugin *plugin)
{
}

GAIM_INIT_PLUGIN(smartear, __init_plugin, info);
